@@ -4,6 +4,7 @@
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2002, 2003 Red Hat, Inc.
  * Copyright (C) 2003 Rob Adams
+ * Copyright (C) 2005 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -369,20 +370,6 @@ avoid_being_obscured_as_second_modal_dialog (MetaWindow *window,
                   "focus window\n",
                   window->desc);
     }
-
-  meta_topic (META_DEBUG_PLACEMENT,
-              "Status:\n"
-              "  denied focus: %d\n"
-              "  modal:        %d\n"
-              "  same app:     %d\n"
-              "  overlaps      %d\n",
-              window->denied_focus_and_not_transient,
-              window->wm_state_modal,
-              meta_window_same_application (window, focus_window),
-              meta_rectangle_intersect (&window->rect,
-                                        &focus_window->rect,
-                                        &overlap)
-              );
 }
 
 static int
@@ -542,16 +529,6 @@ center_tile_rect_in_area (MetaRectangle *rect,
   rect->y = work_area->y + fluff;
 }
 
-static gboolean
-rect_fits_in_work_area (MetaRectangle *work_area,
-                        MetaRectangle *rect)
-{
-  return ((rect->x >= work_area->x) &&
-	  (rect->y >= work_area->y) &&
-	  (rect->x + rect->width <= work_area->x + work_area->width) &&
-	  (rect->y + rect->height <= work_area->y + work_area->height));
-}
-
 /* Find the leftmost, then topmost, empty area on the workspace
  * that can contain the new window.
  *
@@ -610,13 +587,15 @@ find_first_fit (MetaWindow *window,
 
   for (i = 0; i < n_xineramas; i++)
     {
+#ifdef WITH_VERBOSE_MODE
+      char xinerama_location_string[RECT_LENGTH];
+      meta_rectangle_to_string (&window->screen->xinerama_infos[xineramas_list[i]].rect,
+                                xinerama_location_string);
       meta_topic (META_DEBUG_XINERAMA,
-		  "Natural xinerama %d is %d,%d %dx%d\n",
+		  "Natural xinerama %d is %s\n",
 		  i,
-		  window->screen->xinerama_infos[xineramas_list[i]].x_origin,
-		  window->screen->xinerama_infos[xineramas_list[i]].y_origin,
-		  window->screen->xinerama_infos[xineramas_list[i]].width,
-		  window->screen->xinerama_infos[xineramas_list[i]].height);
+		  xinerama_location_string);
+#endif
     }
 
   /* try each xinerama in the natural ordering in turn */
@@ -627,7 +606,7 @@ find_first_fit (MetaWindow *window,
 
       center_tile_rect_in_area (&rect, &work_area);
 
-      if (rect_fits_in_work_area (&work_area, &rect) &&
+      if (meta_rectangle_contains_rect (&work_area, &rect) &&
           !rectangle_overlaps_some_window (&rect, windows))
         {
           *new_x = rect.x;
@@ -655,7 +634,7 @@ find_first_fit (MetaWindow *window,
           rect.x = outer_rect.x;
           rect.y = outer_rect.y + outer_rect.height;
       
-          if (rect_fits_in_work_area (&work_area, &rect) &&
+          if (meta_rectangle_contains_rect (&work_area, &rect) &&
               !rectangle_overlaps_some_window (&rect, below_sorted))
             {
               *new_x = rect.x;
@@ -686,7 +665,7 @@ find_first_fit (MetaWindow *window,
           rect.x = outer_rect.x + outer_rect.width;
           rect.y = outer_rect.y;
      
-          if (rect_fits_in_work_area (&work_area, &rect) &&
+          if (meta_rectangle_contains_rect (&work_area, &rect) &&
               !rectangle_overlaps_some_window (&rect, right_sorted))
             {
               *new_x = rect.x;
@@ -830,7 +809,6 @@ meta_window_place (MetaWindow        *window,
       if (parent)
         {
           int w;
-          MetaRectangle area;
 
           meta_window_get_position (parent, &x, &y);
           w = parent->rect.width;
@@ -849,6 +827,16 @@ meta_window_place (MetaWindow        *window,
           if (fgeom)
             y += fgeom->top_height;
 
+#if 0
+          /* FIXME: If no one has complained about this within a couple
+           * months, it means that the constraints.c rewrite fixed this
+           * correctly and this hack is no longer necessary.  Feel free to
+           * submit a patch removing this #ifdef'd out code.  (Comment
+           * written on 2005-10-28).
+           */
+
+          MetaRectangle area;
+
           /* clip to xinerama of parent*/
           meta_window_get_work_area_current_xinerama (parent, &area);
 
@@ -858,6 +846,7 @@ meta_window_place (MetaWindow        *window,
             y = area.y + area.height - window->rect.height;
           if (x < area.x) x = area.x;
           if (y < area.y) y = area.y;     
+#endif
 
           meta_topic (META_DEBUG_PLACEMENT, "Centered window %s over transient parent\n",
                       window->desc);
@@ -867,6 +856,10 @@ meta_window_place (MetaWindow        *window,
           goto done;
         }
     }
+  
+  meta_screen_get_natural_xinerama_list (window->screen,
+					 &xineramas_list,
+					 &n_xineramas);
 
   /* FIXME UTILITY with transient set should be stacked up
    * on the sides of the parent window or something.
@@ -876,25 +869,25 @@ meta_window_place (MetaWindow        *window,
       window->type == META_WINDOW_MODAL_DIALOG ||
       window->type == META_WINDOW_SPLASHSCREEN)
     {
-      /* Center on screen */
+      /* Center on current xinerama (i.e. on current monitor) */
       int w, h;
 
       /* Warning, this function is a round trip! */
       xi = meta_screen_get_current_xinerama (window->screen);
 
-      w = xi->width;
-      h = xi->height;
+      w = xi->rect.width;
+      h = xi->rect.height;
 
       x = (w - window->rect.width) / 2;
       y = (h - window->rect.height) / 2;
 
-      x += xi->x_origin;
-      y += xi->y_origin;
+      x += xi->rect.x;
+      y += xi->rect.y;
       
       meta_topic (META_DEBUG_PLACEMENT, "Centered window %s on screen %d xinerama %d\n",
                   window->desc, window->screen->number, xi->number);
 
-      goto done;
+      goto done_check_denied_focus;
     }
   
   /* Find windows that matter (not minimized, on same workspace
@@ -928,12 +921,8 @@ meta_window_place (MetaWindow        *window,
   xi = meta_screen_get_current_xinerama (window->screen);
   
   /* "Origin" placement algorithm */
-  x = xi->x_origin;
-  y = xi->y_origin;
-
-  meta_screen_get_natural_xinerama_list (window->screen,
-					 &xineramas_list,
-					 &n_xineramas);
+  x = xi->rect.x;
+  y = xi->rect.y;
 
   if (find_first_fit (window, fgeom, windows,
                       xineramas_list, n_xineramas,
@@ -999,11 +988,11 @@ meta_window_place (MetaWindow        *window,
                                               &workarea);      
       meta_window_get_outer_rect (window, &outer);
       
-      if (outer.width >= workarea.width &&
-          outer.height >= workarea.height)
-        {
-          window->maximize_after_placement = TRUE;
-        }
+      if (outer.width >= workarea.width)
+        window->maximize_horizontally_after_placement = TRUE;
+
+      if (outer.height >= workarea.height)
+        window->maximize_vertically_after_placement = TRUE;
     }
 
  done_check_denied_focus:
@@ -1035,8 +1024,8 @@ meta_window_place (MetaWindow        *window,
           focus_window_list = g_list_prepend (NULL, focus_window);
 
           /* Reset x and y ("origin" placement algorithm) */
-          x = xi->x_origin;
-          y = xi->y_origin;
+          x = xi->rect.x;
+          y = xi->rect.y;
 
           found_fit = find_first_fit (window, fgeom, focus_window_list,
                                       xineramas_list, n_xineramas,
@@ -1070,8 +1059,8 @@ meta_window_place (MetaWindow        *window,
  * would be weird otherwise.
  */
 static GSList*
-get_windows_on_same_workspace (MetaWindow *window,
-                               int        *n_windows)
+get_windows_showing_on_same_screen (MetaWindow *window,
+                                    int        *n_windows)
 {
   GSList *windows;
   GSList *all_windows;
@@ -1088,7 +1077,9 @@ get_windows_on_same_workspace (MetaWindow *window,
     {
       MetaWindow *w = tmp->data;
       
-      if (meta_window_should_be_showing (w) && w != window)
+      if (meta_window_should_be_showing (w) &&
+          w->screen == window->screen       &&
+          w != window)
         {            
           windows = g_slist_prepend (windows, w);
           ++i;
@@ -1105,33 +1096,6 @@ get_windows_on_same_workspace (MetaWindow *window,
   return windows;
 }
 
-static gboolean
-rects_overlap_vertically (const MetaRectangle *a,
-                          const MetaRectangle *b)
-{
-  /* if they don't overlap, then either a is above b
-   * or b is above a
-   */
-  if ((a->y + a->height) < b->y)
-    return FALSE;
-  else if ((b->y + b->height) < a->y)
-    return FALSE;
-  else
-    return TRUE;
-}
-
-static gboolean
-rects_overlap_horizontally (const MetaRectangle *a,
-                            const MetaRectangle *b)
-{
-  if ((a->x + a->width) < b->x)
-    return FALSE;
-  else if ((b->x + b->width) < a->x)
-    return FALSE;
-  else
-    return TRUE;
-}
-
 static void
 get_vertical_edges (MetaWindow *window,
                     int       **edges_p,
@@ -1140,43 +1104,53 @@ get_vertical_edges (MetaWindow *window,
   GSList *windows;
   GSList *tmp;
   int n_windows;
-  int *edges;
-  int i, j;
-  int n_edges;
+  GArray *edges;
+  int edge, i;
   MetaRectangle rect;
   MetaRectangle work_area;
   
-  windows = get_windows_on_same_workspace (window, &n_windows);
+  windows = get_windows_showing_on_same_screen (window, &n_windows);
 
-  i = 0;
-  /* 4 = workspace/screen edges */
-  n_edges = n_windows * 2 + 4 + window->screen->n_xinerama_infos - 1; 
-
-  edges = g_new (int, n_edges);
+  edges = g_array_sized_new (FALSE, FALSE, sizeof (int), 
+                             n_windows * 2 + 4 /* 4 = workspace/screen edges */
+                             + window->screen->n_xinerama_infos - 1
+                             + 2 /* active window edges when in wireframe mode */); 
 
   /* workspace/screen edges */
   meta_window_get_work_area_current_xinerama (window, &work_area);
-
-  edges[i] = work_area.x;
-  ++i;
-  edges[i] = work_area.x + work_area.width;
-  ++i;
-  edges[i] = 0;
-  ++i;
-  edges[i] = window->screen->width;
-  ++i;
-
-  g_assert (i == 4);
+  
+  g_array_append_val (edges, work_area.x);
+  edge = work_area.x + work_area.width;
+  g_array_append_val (edges, edge);
+  edge = 0;
+  g_array_append_val (edges, edge);
+  g_array_append_val (edges, window->screen->rect.width);
 
   /* Now get the xinerama screen edges */
-  for (j = 0; j < window->screen->n_xinerama_infos - 1; j++) {
-    edges[i] = window->screen->xinerama_infos[j].x_origin +
-      window->screen->xinerama_infos[j].width;
-    ++i;
-  }
-  
-  meta_window_get_outer_rect (window, &rect);
-  
+  for (i = 0; i < window->screen->n_xinerama_infos - 1; i++)
+    {
+      edge = window->screen->xinerama_infos[i].rect.x +
+             window->screen->xinerama_infos[i].rect.width;
+
+      g_array_append_val (edges, edge);
+    }
+
+  if (window->display->grab_wireframe_active)
+    {
+      int left_edge, right_edge, top_edge, bottom_edge;
+
+      meta_window_get_xor_rect (window, &window->display->grab_wireframe_rect,
+                                &rect);
+
+      window_get_edges (window, &left_edge, &right_edge, 
+                        &top_edge, &bottom_edge);
+
+      g_array_append_val (edges, left_edge);
+      g_array_append_val (edges, right_edge);
+    }
+  else
+    meta_window_get_outer_rect (window, &rect);
+
   /* get window edges */
   tmp = windows;
   while (tmp != NULL)
@@ -1186,24 +1160,23 @@ get_vertical_edges (MetaWindow *window,
 
       meta_window_get_outer_rect (w, &w_rect);
       
-      if (rects_overlap_vertically (&rect, &w_rect))
+      if (meta_rectangle_vert_overlap (&rect, &w_rect))
         {
-          window_get_edges (w, &edges[i], &edges[i+1], NULL, NULL);
-          i += 2;
+          g_array_append_val (edges, w_rect.x);
+          edge = w_rect.x + w_rect.width;
+          g_array_append_val (edges, edge);
         }
 
       tmp = tmp->next;
     }
-  n_edges = i;
-  
   g_slist_free (windows);
 
   /* Sort */
-  qsort (edges, n_edges, sizeof (int), intcmp);
+  qsort (edges->data, edges->len, sizeof (int), intcmp);
 
-  *edges_p = edges;
-  *n_edges_p = n_edges;
-}  
+  *n_edges_p = edges->len;  
+  *edges_p = (int *) g_array_free (edges, FALSE);
+}
 
 static void
 get_horizontal_edges (MetaWindow *window,
@@ -1213,41 +1186,52 @@ get_horizontal_edges (MetaWindow *window,
   GSList *windows;
   GSList *tmp;
   int n_windows;
-  int *edges;
-  int i, j;
-  int n_edges;
+  GArray *edges;
+  int edge, i;
   MetaRectangle rect;
   MetaRectangle work_area;
   
-  windows = get_windows_on_same_workspace (window, &n_windows);
+  windows = get_windows_showing_on_same_screen (window, &n_windows);
 
-  i = 0;
-  n_edges = n_windows * 2 + 4 + window->screen->n_xinerama_infos - 1; /* 4 = workspace/screen edges */
-  edges = g_new (int, n_edges);
+  edges = g_array_sized_new (FALSE, FALSE, sizeof (int), 
+                             n_windows * 2 + 4 /* 4 = workspace/screen edges */
+                             + window->screen->n_xinerama_infos - 1
+                             + 2 /* active window edges when in wireframe mode */); 
 
   /* workspace/screen edges */
   meta_window_get_work_area_current_xinerama (window, &work_area);
   
-  edges[i] = work_area.y;
-  ++i;
-  edges[i] = work_area.y + work_area.height;
-  ++i;
-  edges[i] = 0;
-  ++i;
-  edges[i] = window->screen->height;
-  ++i;
+  g_array_append_val (edges, work_area.y);
+  edge = work_area.y + work_area.height;
+  g_array_append_val (edges, edge);
+  edge = 0;
+  g_array_append_val (edges, edge);
+  g_array_append_val (edges, window->screen->rect.height);
 
-  g_assert (i == 4);
-  
   /* Now get the xinerama screen edges */
-  for (j = 0; j < window->screen->n_xinerama_infos - 1; j++) {
-    edges[i] = window->screen->xinerama_infos[j].y_origin +
-      window->screen->xinerama_infos[j].height;
-    ++i;
-  }
+  for (i = 0; i < window->screen->n_xinerama_infos - 1; i++) 
+    {
+      edge = window->screen->xinerama_infos[i].rect.y +
+             window->screen->xinerama_infos[i].rect.height;
+      g_array_append_val (edges, edge);
+    }
 
-  meta_window_get_outer_rect (window, &rect);
-  
+  if (window->display->grab_wireframe_active)
+    {
+      int left_edge, right_edge, top_edge, bottom_edge;
+
+      meta_window_get_xor_rect (window, &window->display->grab_wireframe_rect,
+                                &rect);
+
+      window_get_edges (window, &left_edge, &right_edge, 
+                        &top_edge, &bottom_edge);
+
+      g_array_append_val (edges, top_edge);
+      g_array_append_val (edges, bottom_edge);
+    }
+  else
+    meta_window_get_outer_rect (window, &rect);
+ 
   /* get window edges */
   tmp = windows;
   while (tmp != NULL)
@@ -1257,62 +1241,77 @@ get_horizontal_edges (MetaWindow *window,
 
       meta_window_get_outer_rect (w, &w_rect);
       
-      if (rects_overlap_horizontally (&rect, &w_rect))
+      if (meta_rectangle_horiz_overlap (&rect, &w_rect))
         {
-          window_get_edges (w, NULL, NULL, &edges[i], &edges[i+1]);
-          i += 2;
+          g_array_append_val (edges, w_rect.y);
+          edge = w_rect.y + w_rect.height;
+          g_array_append_val (edges, edge);
         }
 
       tmp = tmp->next;
     }
-  n_edges = i;
-  
   g_slist_free (windows);
 
   /* Sort */
-  qsort (edges, n_edges, sizeof (int), intcmp);
+  qsort (edges->data, edges->len, sizeof (int), intcmp);
 
-  *edges_p = edges;
-  *n_edges_p = n_edges;  
+  *n_edges_p = edges->len;  
+  *edges_p = (int *) g_array_free (edges, FALSE);
 }
 
 int
 meta_window_find_next_vertical_edge (MetaWindow *window,
+                                     MetaWindowEdgePosition source_edge_position,
                                      gboolean    right)
 {
-  int left_edge, right_edge;
+  int left_edge, right_edge, source_edge;
   int *edges;
   int i;
   int n_edges;
   int retval;
 
   get_vertical_edges (window, &edges, &n_edges);
-  
-  /* Find next */
-  meta_window_get_position (window, &retval, NULL);
 
-  window_get_edges (window, &left_edge, &right_edge, NULL, NULL);
-  
+  /* Find next */
+  if (window->display->grab_wireframe_active)
+    {
+      MetaRectangle rect;
+
+      meta_window_get_xor_rect (window, &window->display->grab_wireframe_rect,
+                                &rect);
+
+      left_edge = rect.x;
+      right_edge = left_edge + rect.width;
+    }
+  else
+    window_get_edges (window, &left_edge, &right_edge, NULL, NULL);
+
+  switch (source_edge_position)
+    {
+      case META_WINDOW_EDGE_LEFT:
+        source_edge = left_edge;
+        break;
+
+      case META_WINDOW_EDGE_RIGHT:
+        source_edge = right_edge;
+        break;
+
+      case META_WINDOW_EDGE_TOP:
+      case META_WINDOW_EDGE_BOTTOM:
+      default:
+        g_assert_not_reached ();
+    }
+
+  retval = source_edge;
+
   if (right)
     {
       i = 0;
       while (i < n_edges)
         {
-          if (edges[i] > right_edge)
+          if (edges[i] > source_edge)
             {
-              /* This is the one we want, snap right
-               * edge of window to edges[i]
-               */
               retval = edges[i];
-              if (window->frame)
-                {
-                  retval -= window->frame->rect.width;
-                  retval += window->frame->child_x;
-                }
-              else
-                {
-                  retval -= window->rect.width;
-                }
               break;
             }
           
@@ -1326,13 +1325,9 @@ meta_window_find_next_vertical_edge (MetaWindow *window,
         {
           --i;
           
-          if (edges[i] < left_edge)
+          if (edges[i] < source_edge)
             {
-              /* This is the one we want */
               retval = edges[i];
-              if (window->frame)
-                retval += window->frame->child_x;
-
               break;
             }
         }
@@ -1346,9 +1341,10 @@ meta_window_find_next_vertical_edge (MetaWindow *window,
 
 int
 meta_window_find_next_horizontal_edge (MetaWindow *window,
+                                       MetaWindowEdgePosition source_edge_position,
                                        gboolean    down)
 {
-  int top_edge, bottom_edge;
+  int top_edge, bottom_edge, source_edge;
   int *edges;
   int i;
   int n_edges;
@@ -1357,30 +1353,45 @@ meta_window_find_next_horizontal_edge (MetaWindow *window,
   get_horizontal_edges (window, &edges, &n_edges);
   
   /* Find next */
-  meta_window_get_position (window, NULL, &retval);
+  if (window->display->grab_wireframe_active)
+    {
+      MetaRectangle rect;
+
+      meta_window_get_xor_rect (window, &window->display->grab_wireframe_rect,
+                                &rect);
+
+      top_edge = rect.y;
+      bottom_edge = top_edge + rect.height;
+    }
+  else
+    window_get_edges (window, NULL, NULL, &top_edge, &bottom_edge);
+
+  switch (source_edge_position)
+    {
+      case META_WINDOW_EDGE_TOP:
+        source_edge = top_edge;
+        break;
+
+      case META_WINDOW_EDGE_BOTTOM:
+        source_edge = bottom_edge;
+        break;
+
+      case META_WINDOW_EDGE_LEFT:
+      case META_WINDOW_EDGE_RIGHT:
+      default:
+        g_assert_not_reached ();
+    }
+
+  retval = source_edge;
 
-  window_get_edges (window, NULL, NULL, &top_edge, &bottom_edge);
-  
   if (down)
     {
       i = 0;
       while (i < n_edges)
         {
-          if (edges[i] > bottom_edge)
+          if (edges[i] > source_edge)
             {
-              /* This is the one we want, snap right
-               * edge of window to edges[i]
-               */
               retval = edges[i];
-              if (window->frame)
-                {
-                  retval -= window->frame->rect.height;
-                  retval += window->frame->child_y;
-                }
-              else
-                {
-                  retval -= window->rect.height;
-                }
               break;
             }
           
@@ -1394,13 +1405,9 @@ meta_window_find_next_horizontal_edge (MetaWindow *window,
         {
           --i;
           
-          if (edges[i] < top_edge)
+          if (edges[i] < source_edge)
             {
-              /* This is the one we want */
               retval = edges[i];
-              if (window->frame)
-                retval += window->frame->child_y;
-
               break;
             }
         }
@@ -1435,11 +1442,10 @@ meta_window_find_nearest_vertical_edge (MetaWindow *window,
       int left_pos, right_pos;
 
       left_pos = edges[i];
-      if (window->frame)
-        left_pos += window->frame->child_x;
 
       if (window->frame)
         {
+          left_pos += window->frame->child_x;
           right_pos = edges[i] - window->frame->rect.width;
           right_pos += window->frame->child_x;
         }
@@ -1506,11 +1512,10 @@ meta_window_find_nearest_horizontal_edge (MetaWindow *window,
       int top_pos, bottom_pos;
 
       top_pos = edges[i];
-      if (window->frame)
-        top_pos += window->frame->child_y;
 
       if (window->frame)
         {
+          top_pos += window->frame->child_y;
           bottom_pos = edges[i] - window->frame->rect.height;
           bottom_pos += window->frame->child_y;
         }
