@@ -3,6 +3,7 @@
 /* 
  * Copyright (C) 2001 Havoc Pennington
  * Copyright (C) 2002 Red Hat, Inc.
+ * Copyright (C) 2005 Elijah Newren
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -205,7 +206,7 @@ meta_ui_tab_popup_new (const MetaTabEntry *entries,
       if (entries[i].title)
         {
           gchar *tmp;
-          if (entries[i].minimized)
+          if (entries[i].hidden)
             {
               tmp = g_strdup_printf ("[%s]", entries[i].title);
             }
@@ -234,21 +235,21 @@ meta_ui_tab_popup_new (const MetaTabEntry *entries,
       if (te->icon)
         {
           g_object_ref (G_OBJECT (te->icon));
-          if (entries[i].minimized)
+          if (entries[i].hidden)
             te->dimmed_icon = dimm_icon (entries[i].icon);
         }
 
       if (outline)
         {
-          te->rect.x = entries[i].x;
-          te->rect.y = entries[i].y;
-          te->rect.width = entries[i].width;
-          te->rect.height = entries[i].height;
+          te->rect.x = entries[i].rect.x;
+          te->rect.y = entries[i].rect.y;
+          te->rect.width = entries[i].rect.width;
+          te->rect.height = entries[i].rect.height;
 
-          te->inner_rect.x = entries[i].inner_x;
-          te->inner_rect.y = entries[i].inner_y;
-          te->inner_rect.width = entries[i].inner_width;
-          te->inner_rect.height = entries[i].inner_height;
+          te->inner_rect.x = entries[i].inner_rect.x;
+          te->inner_rect.y = entries[i].inner_rect.y;
+          te->inner_rect.width = entries[i].inner_rect.width;
+          te->inner_rect.height = entries[i].inner_rect.height;
         }
 
       tab_entries = g_list_prepend (tab_entries, te);
@@ -738,7 +739,8 @@ selectable_workspace_new (MetaWorkspace *workspace)
   
   widget = g_object_new (meta_select_workspace_get_type (), NULL);
 
-  screen_aspect = (double) workspace->screen->height / (double) workspace->screen->width;
+  screen_aspect = (double) workspace->screen->rect.height /
+                  (double) workspace->screen->rect.width;
   
   /* account for select rect */ 
   gtk_widget_set_size_request (widget,
@@ -808,6 +810,40 @@ meta_select_workspace_class_init (MetaSelectWorkspaceClass *klass)
   widget_class->expose_event = meta_select_workspace_expose_event;
 }
 
+/**
+ * meta_convert_meta_to_wnck() converts a MetaWindow to a
+ * WnckWindowDisplayInfo window that is used to build a thumbnail of a
+ * workspace.
+ **/
+static WnckWindowDisplayInfo
+meta_convert_meta_to_wnck (MetaWindow *window, MetaScreen *screen)
+{
+  WnckWindowDisplayInfo wnck_window;
+  wnck_window.icon = window->icon;
+  wnck_window.mini_icon = window->mini_icon;
+  
+  wnck_window.is_active = FALSE;
+  if (window == window->display->expected_focus_window)
+    wnck_window.is_active = TRUE;
+
+  if (window->frame)
+    {
+      wnck_window.x = window->frame->rect.x;
+      wnck_window.y = window->frame->rect.y;
+      wnck_window.width = window->frame->rect.width;
+      wnck_window.height = window->frame->rect.height;
+    }
+  else
+    {
+      wnck_window.x = window->rect.x;
+      wnck_window.y = window->rect.y;
+      wnck_window.width = window->rect.width;
+      wnck_window.height = window->rect.height;
+    }
+  return wnck_window;
+}
+
+
 static gboolean
 meta_select_workspace_expose_event (GtkWidget      *widget,
                                     GdkEventExpose *event)
@@ -836,7 +872,7 @@ meta_select_workspace_expose_event (GtkWidget      *widget,
                           workspace != workspace->screen->active_workspace;
 
       if (window->skip_pager || 
-          window->minimized || 
+          !meta_window_showing_on_its_workspace (window) ||
           window->unmaps_pending ||
           ignoreable_sticky)
         {
@@ -844,26 +880,7 @@ meta_select_workspace_expose_event (GtkWidget      *widget,
         }
       else
         {
-          windows[i].icon = window->icon;
-          windows[i].mini_icon = window->mini_icon;
-          windows[i].is_active = 
-            (window == window->display->expected_focus_window);
-
-          if (window->frame)
-            {
-              windows[i].x = window->frame->rect.x;
-              windows[i].y = window->frame->rect.y;
-              windows[i].width = window->frame->rect.width;
-              windows[i].height = window->frame->rect.height;
-            }
-          else
-            {                
-              windows[i].x = window->rect.x;
-              windows[i].y = window->rect.y;
-              windows[i].width = window->rect.width;
-              windows[i].height = window->rect.height;
-            }
-
+          windows[i] = meta_convert_meta_to_wnck (window, workspace->screen);
           i++;
         }
       tmp = tmp->next;
@@ -877,8 +894,8 @@ meta_select_workspace_expose_event (GtkWidget      *widget,
                        SELECT_OUTLINE_WIDTH,
                        widget->allocation.width - SELECT_OUTLINE_WIDTH * 2,
                        widget->allocation.height - SELECT_OUTLINE_WIDTH * 2,
-                       workspace->screen->width,
-                       workspace->screen->height,
+                       workspace->screen->rect.width,
+                       workspace->screen->rect.height,
                        NULL,
                        (workspace->screen->active_workspace == workspace),
                        windows,
